/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.dsl

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 5/25/11
 * Time: 10:49 AM
 */
enum ArtifactType {
    jobs,
    bindings,
    propertyFiles,
    schemas;

    def validate(config) {

        def instanceOfList = {propertyValue -> propertyValue instanceof List}

        switch (this) {
            case jobs:
            case bindings:
                def validator = {List propertyValue ->
                    return isListOfClasses(propertyValue)
                }
                doValidate(config, instanceOfList, validator)
                break
            case schemas:

                def instanceOf = {propertyValue -> propertyValue instanceof Map}
                def validator = {propertyValue ->
                    def contentIsRightType = true
                    propertyValue.values().each {schemaList ->
                        if (schemaList instanceof List) {
                            contentIsRightType = isListOfClasses(schemaList)
                        } else {
                            contentIsRightType = false
                        }
                    }

                    return contentIsRightType
                }

                doValidate(config, instanceOf, validator)
                break
            case propertyFiles:
                def validator = {List propertyValue ->
                    validateList(propertyValue) {listItem ->
                        return listItem instanceof String
                    }
                }
                doValidate(config, instanceOfList, validator)
                break
            default:
                throw new RuntimeException("artifact ${this} is not supported")
        }
    }

    private boolean isListOfClasses(List list) {
        validateList(list) {listItem ->
            return listItem instanceof Class
        }
    }

    private boolean validateList(List list, Closure correctType) {
        def contentIsCorrect = true
        list.each {
            if (!correctType(it)) {
                contentIsCorrect = false
            }
        }

        return contentIsCorrect
    }

    String getTypeDescription() {
        switch (this) {
            case jobs:
            case bindings:
                return "list of classes"
            case schemas:
                return "map of lists of classes"
            default:
                return "unknown type for ${this.toString()}"
        }
    }

    def doValidate(config, Closure instanceOf, Closure validator) {
        def hasProperty
        def propertyValue

        if (config instanceof Map) {
            hasProperty = config.containsKey(this.toString())
            if (hasProperty) {
                propertyValue = config[this.toString()]
            }
        } else {
            def meta = config.metaClass
            if (meta == null) {
                throw new IllegalArgumentException("cannot extract property information from config ${config}")
            }
            hasProperty = meta.hasProperty(config, this.toString())
            if (hasProperty) {
                propertyValue = hasProperty.getProperty(config)
            }
        }

        if (hasProperty) {
            def boolean valueIsRightType = instanceOf(propertyValue)
            def boolean contentIsRightType = true
            if (valueIsRightType) {
                contentIsRightType = validator.call(propertyValue)
            }

            if (!valueIsRightType || !contentIsRightType) {
                throw new IllegalConfigurationException(artifactType: this.toString(), actualContent: propertyValue)
            }
        }
    }
}

