/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.dsl

import ch.qos.logback.classic.Level
import ch.qos.logback.classic.Logger
import metridoc.dsl.binding.BindingLoader
import metridoc.extension.ExtensionRunner
import metridoc.thread.IsolatedThreadGroup
import metridoc.utils.PropertyUtils
import metridoc.utils.SchemaUtils
import org.slf4j.LoggerFactory
import metridoc.utils.Assert

/**
 * Primary interface to run jobs.  Useful for embedding metridoc into standard java apps as well.  For
 * {@link JobBuilder} documentation see http://code.google.com/p/metridoc/wiki/JobBuilder, for documentation on job
 * files please see http://code.google.com/p/metridoc/wiki/JobFile
 *
 *
 */
class JobBuilder {

    def ExtensionRunner extensionRunner
    def Binding services = new Binding()
    def JobManager manager
    private static JobBuilder DEFAULT_JOB_BUILDER
    def log = LoggerFactory.getLogger("${METRIDOC_JOBS}.builder")
    boolean stopCalled = false
    public static final JOBS = ArtifactType.jobs.toString()
    public static final BINDINGS = ArtifactType.bindings.toString()
    public static final SCHEMAS = ArtifactType.schemas.toString()
    public static final JOB_PROPERTIES = ArtifactType.propertyFiles.toString()
    public static final String DEFAULT_JOB = "defaultJob"
    public static final String METRIDOC_JOBS = "metridoc.jobs"
    public static final String METRIDOC_BUILDER = "${METRIDOC_JOBS}.builder"



    ExtensionRunner getExtensionRunner() {
        if (extensionRunner) {
            return extensionRunner
        }

        extensionRunner = new ExtensionRunner(jobManager: getJobManager())
    }

    def JobBuilder load(String... jobFiles) {
        getJobManager().parseJobs jobFiles
        return this
    }

    def JobBuilder load(Class... jobClasses) {
        log.debug("loading jobs ${jobClasses}")
        getJobManager().parseJobs jobClasses
        return this
    }

    def JobManager getJobManager() {
        if (manager == null) {
            manager = new JobManager(services: services)
        }

        return manager
    }

    def void run() {
        log.debug("no job specified, so '${DEFAULT_JOB}' will be run")
        run(new LinkedHashMap(), DEFAULT_JOB)
    }

    def void run(LinkedHashMap args = null, String... jobs) {
        configureBuilder this, args
        getJobManager().runJobs jobs
    }

    def JobBuilder useTracing(boolean useTracing) {
        getJobManager().camelContext.setTracing useTracing
    }

    def JobBuilder loadConfig(config) {
        validateConfig(config)
        loadConfigBean(config)

        return this
    }

    private void loadConfigBean(config) {
        if (hasBindings(config)) {
            loadBindings(config.bindings)
        }
        if (hasJobs(config)) {
            int size = config.jobs.size()
            load(config.jobs.toArray(new Class[size]))
        }
        if (hasProperties(config)) {
            int size = config.propertyFiles.size()
            loadProperties(config.propertyFiles.toArray(new String[size]))
        }
        if (hasSchemas(config)) {
            loadSchemas(config.schemas)
        }
    }

    private void loadSchemas(LinkedHashMap<String, List<Class>> schemas) {
        def services = getServices()
        def schemaUtils = new SchemaUtils()

        schemas.keySet().each {db ->
            def dbParams = services[db]
            Assert.notNull(dbParams, "The datasource service ${db} does not exist")
            schemaUtils.jdbcUrl = dbParams.jdbcUrl
            schemaUtils.dbArgs = dbParams
            schemas.db.each {schema ->
                schemaUtils.runSchema(schema)
            }
        }

    }

    private boolean hasSchemas(config) {
        has(SCHEMAS, config)
    }

    JobBuilder loadProperties(String... properties) {
        def util = new PropertyUtils()
        properties.each {String propertyFile ->
            def config = util.getConfig(propertyFile)
        }
    }

    private boolean hasProperties(config) {
        has(JOB_PROPERTIES, config)
    }

    private boolean hasBindings(config) {
        has(BINDINGS, config)
    }

    def JobBuilder loadBindings(List<Class> bindings) {
        bindings.each {Class binding ->
            loadBinding(binding)
        }

        return this
    }

    private boolean hasJobs(config) {
        has(JOBS, config)
    }

    private boolean has(String type, config) {
        if (config instanceof Map) {
            return config.containsKey(type)
        } else {
            def meta = config.metaClass
            if (meta) {
                return meta.hasProperty(config, type) != null
            } else {
                throw new IllegalConfigurationException(message: "Could not get meta class from configuration " +
                        "${config}, can't tell if it has property ${type}")
            }
        }
    }

    def JobBuilder loadBinding(binding) {
        BindingLoader.loadBinding(binding, this)
        return this
    }

    private void setLoggingLevel(Level logLevel) {
        Logger logger = (Logger) LoggerFactory.getLogger(Logger.ROOT_LOGGER_NAME)
        logger.setLevel(logLevel)
    }

    def static void run(LinkedHashMap args = null, Closure closure) {
        run(args, DEFAULT_JOB, closure)
    }

    def static void run(LinkedHashMap args = null, String jobName, Closure closure) {
        JobBuilder.run(args, [jobName] as String[], closure)
    }

    def static void run(LinkedHashMap args = null, String[] jobNames, Closure closure) {
        IsolatedThreadGroup.run(name: "${JobBuilder.METRIDOC_JOBS}.builder" as String) {

            def builder = new JobBuilder()
            try {
                configureBuilder builder, args
                def jobManager = builder.getJobManager()
                if (closure) {
                    jobManager.loadJobs closure
                }
                jobManager.runJobs jobNames
            } finally {
                builder.stop()
            }
        }
    }

    def static void jobs(LinkedHashMap args = null, Closure closure) {
        jobs(args, DEFAULT_JOB, closure)
    }

    def static void jobs(LinkedHashMap args = null, String jobName, Closure closure) {
        run(args, jobName, closure)
    }

    def static isJob(Object job) {
        def manager = getDefaultJobBuilder().getJobManager()
        job.metaClass.mixin JobRunner

        job.services = manager.services
        if(job instanceof Script) {
            job.services = job.binding
            job.services["out"] = System.out
        }

        return job
    }



    static JobBuilder getDefaultJobBuilder() {
        if(DEFAULT_JOB_BUILDER) {
            return DEFAULT_JOB_BUILDER
        }

        DEFAULT_JOB_BUILDER = new JobBuilder()
        configureBuilder(DEFAULT_JOB_BUILDER, null)
        return DEFAULT_JOB_BUILDER
    }


    def void stop() {
        extensionRunner.onStop.each {Closure onStop ->
            onStop.call(this)
        }
        stopCalled = true
    }

    private static void configureBuilder(JobBuilder builder, args) {
        builder.extensionRunner.binding.setVariable("jobBuilder", builder)
        builder.extensionRunner.runExtensions()
        builder.extensionRunner.jobBuilderConfigurations.each {Closure config ->
            config.call(args, builder)
        }

        Runtime.runtime.addShutdownHook(
            new Thread({
                if(!builder.stopCalled) {
                    builder.stop()
                }
            })
        )
    }

    private validateConfig(config) {
        ArtifactType.values().each {artifact ->
            artifact.validate(config)
        }
    }
}
