/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package metridoc.extension

import metridoc.utils.IOUtils;
import metridoc.dsl.JobManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory
import metridoc.dsl.JobBuilder

/**
 *
 * @author Tommy Barker
 */
class ExtensionRunner {
    static final String DEFAULT_EXTENSION = "MetridocExtension.groovy";
    private static final Logger LOG = LoggerFactory.getLogger(ExtensionRunner.class);
    private ClassLoader classLoader;
    Binding binding = new Binding();
    private static boolean extensionsHaveRun = false;
    JobManager jobManager = new JobManager();
    List<Closure> jobBuilderConfigurations = []
    List<Closure> onStop = []
    JobBuilder jobBuilder
    boolean staticRunner = false
    LinkedHashMap args

    public static ExtensionRunner getExtensionRunner() {
        return getExtensionRunner(null)
    }

    public static ExtensionRunner getExtensionRunner(LinkedHashMap args) {
        def extensionRunner = new ExtensionRunner(jobBuilder: new JobBuilder(), staticRunner: true, args: args)

        extensionRunner.jobManager = extensionRunner.jobBuilder.jobManager
        if (!extensionsHaveRun) {
            extensionRunner.runExtensions(DEFAULT_EXTENSION)
        }

        return extensionRunner;
    }

    public ClassLoader getClassLoader() {
        if (classLoader == null) {
            classLoader = ExtensionRunner.class.getClassLoader();
        }

        return classLoader;
    }

    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    void runExtensions() throws Exception {
        runExtensions(DEFAULT_EXTENSION)
    }

    void runExtensions(String... extensions) throws Exception {
        runExtensions(binding, extensions);
    }

    synchronized void runExtensions(Binding binding, String... extensions) throws Exception {
        for (String extension : extensions) {
            LOG.debug("searching for extension {}", extension);
            List<InputStream> resources = IOUtils.getResources(extension, getClassLoader());
            if (resources.isEmpty()) {
                LOG.warn("could not find any extensions for {}", extension);
            }

            for (InputStream inputStream : resources) {
                GroovyShell shell = new GroovyShell(getClassLoader(), binding);
                Script script = shell.parse(new InputStreamReader(inputStream), extension);
                script.metaClass.configureJobBuilder = {Closure closure ->
                    jobBuilderConfigurations.add(closure)
                }
                script.metaClass.onStop = {Closure closure ->
                    onStop.add(closure)
                }
                script.run()
            }

            if(staticRunner) {
                jobBuilderConfigurations.each {Closure config ->
                    config.call(args, jobBuilder)
                }
            }
        }

        Map variables = binding.getVariables();
        for (Object key : variables.keySet()) {
            jobManager.getServices().setVariable(key, variables.get(key))
        }
    }
}

