/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.thread

import org.apache.commons.lang.RandomStringUtils
import org.slf4j.LoggerFactory

/**
 * Based on the IsolatedThreadGroup from the maven exec plugin
 *
 * a ThreadGroup to isolate execution and collect exceptions.
 */

class IsolatedThreadGroup extends ThreadGroup {
    Throwable uncaughtException; // synchronize access to this
    final static String DEFAULT_NAME = "metridoc.jobs.isolatedThreadGroup"
    final static long ONE_SECOND = 1000
    final static long ONE_MINUTE = ONE_SECOND * 60
    final static long FIVE_SECONDS = ONE_SECOND * 5
    final static long FIFTEEN_MINUTES = ONE_MINUTE * 15
    def log

    public IsolatedThreadGroup(String name) {
        super(name);
    }

    Throwable getUncaughtException() {
        if(parentHasUnCaughtException()) {
            return parent.uncaughtException
        }

        return uncaughtException
    }

    private boolean parentHasUnCaughtException() {
        if(parentInstanceOfIsolatedThreadGroup()) {
            if(parent.uncaughtException) {
                return true
            }
        }
    }

    def getLog() {
        if (log) {
            return log
        }

        log = LoggerFactory.getLogger(getName())
    }

    synchronized void uncaughtException(Thread thread, Throwable throwable) {
        if (throwable instanceof ThreadDeath) {
            return //harmless
        }
        boolean additionalException = false;

        if(parentInstanceOfIsolatedThreadGroup()) {
            parent.uncaughtException(thread, throwable)
            return //roll it up
        }

        if (getUncaughtException() == null) // only remember the first one
        {
            uncaughtException = throwable; // will be reported eventually
        }
        else if (throwable != uncaughtException) { //only report it if it is different, otherwise ignore
            additionalException = true
        }
        if (additionalException) {
            handleAdditionalException(thread, throwable)
        }
    }

    private boolean parentInstanceOfIsolatedThreadGroup() {
        parent && parent instanceof IsolatedThreadGroup
    }

    def static void run(LinkedHashMap args = null, Closure closure) {

        def config = new ThreadGroupConfig(args)
        def threadGroup = new IsolatedThreadGroup(config.name)

        if (config.log) {
            threadGroup.log = args.log
        }

        def log = threadGroup.getLog()
        def runnable = new Runnable() {
            void run() {
                closure()
            }
        }
        def bootStrapThread = new Thread(threadGroup, runnable, config.name)
        bootStrapThread.start()


        if (config.joinSpawnedThreads) {

            threadGroup.joinThreads(config.timeout) {
                threadGroup.getNonDaemonThreads()
            }
            if (config.includeDaemon) {
                threadGroup.joinThreads(config.timeout) {
                    return threadGroup.getAllActiveThreads()
                }

                threadGroup.destroy() //all threads should be stopped, so let's destroy it
            }
        } else {
            threadGroup.joinThreads(config.timeout, {bootStrapThread.alive ? bootStrapThread : null})
        }

        if (threadGroup.uncaughtException) {
            throw threadGroup.uncaughtException
        }
    }

    boolean activeNonDaemonThreads() {
        return getNonDaemonThreads().size() > 0
    }

    private joinThreads(long timeout, Closure retrieveThreads) {
        /*
            since threads can be spawned by threads we must check for active threads until there are none
        */
        while (retrieveThreads()) {
            retrieveThreads().each {Thread thread ->
                getLog().debug "joining ${thread}"
                thread.join(timeout)
                getLog().debug "done joining ${thread}"
                if (thread.alive) {
                    handleSuspiciousThread(thread, timeout)
                }
            }
        }
    }

    List<Thread> getAllActiveThreads() {
        def result = []
        def threads = new Thread[activeCount()]
        enumerate(threads)
        threads.each {thread ->
            if (thread && thread.alive) {
                result.add(thread)
            }
        }
        return result
    }

    List<Thread> getNonDaemonThreads() {
        def result = []
        getAllActiveThreads().each {Thread thread ->
            if (thread.alive && !thread.daemon) {
                result.add(thread)
            }
        }

        return result
    }

    /**
     * essentially an extension point to provide custom logic to handle threads that refuse to quit.  Right now
     * suspicious thread are just logged which could lead to threads never terminating
     *
     * @param thread
     * @param timeout
     */
    protected void handleSuspiciousThread(Thread thread, long timeout) {
        def message = "despite waiting at least ${timeout} milliseconds, the thread ${thread} has not " +
                "completed.  It is possible this thread will never complete or needs more time to complete"
        getLog().warn message
    }

    /**
     * The thread group only handles the first exception.  This extension allows custom handling of subsequent
     * exceptions.  The default behaviour just logs it as an error
     * @param thread
     * @param throwable
     */
    protected void handleAdditionalException(Thread thread, Throwable throwable) {
        String message = "An additional exception in thread group ${this.name} from thread ${thread} occurred"
        getLog().error message, throwable
    }
}

class ThreadGroupConfig {
    String name = "${IsolatedThreadGroup.DEFAULT_NAME}_${RandomStringUtils.randomAlphanumeric(5)}"
    boolean includeDaemon = false
    long timeout = IsolatedThreadGroup.FIFTEEN_MINUTES
    boolean joinSpawnedThreads = false
    def log
}