/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.utils

import javax.sql.DataSource
import liquibase.integration.spring.SpringLiquibase
import metridoc.sql.SqlLocator
import metridoc.schema.loader.GroovyClassSchemaResource
import org.springframework.core.io.DefaultResourceLoader
import com.mchange.v2.c3p0.ComboPooledDataSource

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 4/29/11
 * Time: 12:50 PM
 * To change this template use File | Settings | File Templates.
 */
class SchemaUtils {

    def String jdbcUrl
    def DataSource dataSource
    def SpringLiquibase springLiquibase
    def dbArgs;
    private static final NO_PROPERTY_ERROR_PREFIX = "The schema util MUST have"
    private static final NO_PROPERTY_ERROR_POSTFIX = "property set"
    private static final JDBC_URL = "jdbcUrl"

    def void addNotNullConstraints(String tableName, Map<String, String> columns) {
        columns.each {column, dataType ->
            addNotNullConstraint(tableName: tableName, columnName: column, columnDataType: dataType)
        }
    }

    def void createBasicTable(String tableName, Map<String, String> columns) {
        createTable(tableName: tableName) {
            column(name: "${tableName}_id", type: "BIGINT UNSIGNED", autoIncrement: true) {
                constraints(nullable: false, primaryKey: true)
            }
            columns.each {name, type ->
                column(name: name, type: type)
            }
            column(name: "load_time", type: "TIMESTAMP", defaultValueComputed: "CURRENT_TIMESTAMP") {
                constraints(nullable: false)
            }
        }
    }

    /**
     * creates a batch of single valued tables where the column name is equal to the table name
     *
     * @param columns
     */
    def void createOneColumnTables(Map<String, String> columns) {
        columns.each {String column, String type ->
            createBasicTable(column, [column:type])
        }
    }

    def void runSchema(schema) {

        def liquibase = getSpringLiquibase()
        if (schema instanceof Class) {
            liquibase.resourceLoader = new GroovyClassSchemaResource()
            liquibase.changeLog = "${schema.name}.xml" //adding the xml extension tricks liquibase into using the xml parser
        } else {
            liquibase.changeLog = schema
            liquibase.resourceLoader = new DefaultResourceLoader()
        }

        liquibase.contexts = StringUtils.EMPTY
        liquibase.afterPropertiesSet()
    }

    SpringLiquibase getSpringLiquibase() {
        if(springLiquibase) {
            return springLiquibase
        }

        springLiquibase = new SpringLiquibase()
        springLiquibase.dataSource = getDataSource()

        return springLiquibase
    }

    DataSource getDataSource() {
        if(dataSource) {
            return dataSource
        }

        if(!dbArgs.driverClass) {
            dbArgs.driverClass = getDriverClass()
        }

        return new ComboPooledDataSource(dbArgs)
    }

    String getDriverClass() {
        SqlLocator.getProvider(getJdbcUrl()).driverClass.name
    }

    String getJdbcUrl() {
        if(jdbcUrl) {
            return jdbcUrl
        }

        handleMissingProperty(JDBC_URL)
    }

    private handleMissingProperty(String propertyName) {
        throw new IllegalStateException("${NO_PROPERTY_ERROR_PREFIX} ${propertyName} ${NO_PROPERTY_ERROR_POSTFIX}")
    }
}
